/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee.test;

import com.cloudera.keytrustee.TrusteeKeyProvider;
import com.cloudera.keytrustee.TrusteeKeyProviderConfiguration;
import com.google.common.base.Preconditions;
import org.apache.hadoop.crypto.key.KeyProviderFactory;
import org.apache.hadoop.crypto.key.kms.server.KMSConfiguration;
import org.junit.Assert;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.KeyProvider;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.StandardCopyOption;
import java.security.NoSuchAlgorithmException;
import java.util.List;
import java.util.Map;
import java.util.HashMap;

public class ClientTest extends TestHelper {

  private static final String KEYTRUSTEE_SERVER = "keytrustee.server";

  //private static final String KEYTRUSTEE_SERVER_DEFAULT = "192.168.2.215";
  //private static final String KEYTRUSTEE_SERVER_DEFAULT = "50.112.136.40"; // ztrustee demo server
  //private static final String KEYTRUSTEE_SERVER_DEFAULT = "ec2-54-196-159-5.compute-1.amazonaws.com"; // Hector's active server
  //private static final String KEYTRUSTEE_SERVER_DEFAULT = "172.26.0.198"; // Anthony's test server
  private static final String KEYTRUSTEE_SERVER_DEFAULT = "ayoungga-keytrustee.ent.cloudera.com"; // Anthony's test server

  private static final String SSL_INSECURE = "ssl.insecure";
  private static final String SSL_INSECURE_DEFAULT = "true";

  private static final String REG_ORG = "register.org";
  private static final String REG_ORG_DEFAULT = "kptest";
  //private static final String REG_ORG_DEFAULT = "org";

  private static final String REG_AUTH = "register.auth";
  private static final String REG_AUTH_DEFAULT = "numeroparadaindicaproceso";
  //private static final String REG_AUTH_DEFAULT = "org";

  @Rule
  public TemporaryFolder folder = new TemporaryFolder();

  protected String getServer() {
    return System.getProperty(KEYTRUSTEE_SERVER, KEYTRUSTEE_SERVER_DEFAULT);
  }

  protected String getOrg(){
    return System.getProperty(REG_ORG, REG_ORG_DEFAULT);
  }

  protected String getAuth(){
    return System.getProperty(REG_AUTH, REG_AUTH_DEFAULT);
  }

  protected boolean getSslInsecure() {
    return Boolean.parseBoolean(System.getProperty(SSL_INSECURE, SSL_INSECURE_DEFAULT));
  }

  @Test
  public void testNewProvider() {
    registerNewProvider();
  }

  @Test
  public void testCreateKey() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      byte[] material = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      String name = "testCreateKey-" + getUniqueSuffix();

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion = tkp.createKey(name, material, options);

      String outputName = keyVersion.getName();
      byte[] outputMaterial = keyVersion.getMaterial();
      String outputVersionName = keyVersion.getVersionName();

      Assert.assertEquals("Key name returned from provider doesn't match input name.", name, outputName);
      Assert.assertEquals("Key material length returned from provider doesn't match input material length.", material.length, outputMaterial.length);
      Assert.assertArrayEquals("Key material returned from provider doesn't match input material.", material, outputMaterial);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }

  }

  @Test
  public void testGetKeyVersion() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      byte[] material = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      String name = "testGetKeyVersion-" + getUniqueSuffix();

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersionExpected = tkp.createKey(name, material, options);

      String versionName = keyVersionExpected.getVersionName();

      KeyProvider.KeyVersion keyVersionActual = tkp.getKeyVersion(versionName);

      String outputName = keyVersionActual.getName();
      byte[] outputMaterial = keyVersionActual.getMaterial();

      Assert.assertEquals("Key name returned from provider doesn't match input name.", name, outputName);
      Assert.assertEquals("Key material length returned from provider doesn't match input material length.", material.length, outputMaterial.length);
      Assert.assertArrayEquals("Key material returned from provider doesn't match input material.", material, outputMaterial);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testGetKeyVersions() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testGetKeyVersions-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion(name, material2);

      List<KeyProvider.KeyVersion> keyVersionsActual = tkp.getKeyVersions(name);

      Assert.assertEquals("Key versions list is not of expected size.", 3, keyVersionsActual.size());
      Assert.assertArrayEquals("First key material from provider doesn't match last material stored.", material0, keyVersionsActual.get(2).getMaterial());
      Assert.assertArrayEquals("Second key material from provider doesn't match second material stored.", material1, keyVersionsActual.get(1).getMaterial());
      Assert.assertArrayEquals("Third key material from provider doesn't match first material stored.", material2, keyVersionsActual.get(0).getMaterial());
    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testRollNewVersion() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testRollNewVersion-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      Assert.assertArrayEquals("Key material from provider doesn't match material of last key version.", material1, keyVersion1.getMaterial());

      try {
        KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion("noname", material1);
        Assert.fail("Roll key succeeded when non-existent name was passed as parameter.");
      } catch (IOException e) {
        // expected;
      }

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testRollNewVersionOneArg() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testRollNewVersionOneArg-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name);

      try {
        KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion("noname");
        Assert.fail("Roll key succeeded when non-existent name was passed as parameter.");
      } catch (IOException e) {
        // expected;
      }

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testGetCurrentKey() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testGetCurrentKey-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion(name, material2);

      KeyProvider.KeyVersion keyVersionActual = tkp.getCurrentKey(name);

      Assert.assertArrayEquals("Key material from getCurrentKey doesn't match material of last key version.", material2, keyVersionActual.getMaterial());
    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testGetMetadata() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      byte[] material = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      String name = "testGetMetadata-" + getUniqueSuffix();

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      HashMap<String,String> attributes = getDefaultAttributes();
      options.setAttributes(attributes);

      String description = "test description";
      options.setDescription(description);

      KeyProvider.KeyVersion keyVersionExpected = tkp.createKey(name, material, options);

      String versionName = keyVersionExpected.getVersionName();

      KeyProvider.Metadata actualMetadata = tkp.getMetadata(name);

      Assert.assertEquals("Ciphers do not match.", conf.get(KeyProvider.DEFAULT_CIPHER_NAME), actualMetadata.getCipher());
      Assert.assertTrue("Bit lengths do not match.", Integer.parseInt(conf.get(KeyProvider.DEFAULT_BITLENGTH_NAME)) == actualMetadata.getBitLength());
      Assert.assertEquals("Descriptions do not match.", description, actualMetadata.getDescription());

      Map<String, String> actualAttributes = actualMetadata.getAttributes();
      Assert.assertEquals("Attribute map sizes do not match.", attributes.size(), actualAttributes.size());

      for (String key : attributes.keySet()) {
        String expectedValue = attributes.get(key);
        String actualValue= actualAttributes.get(key);
        Assert.assertEquals("Attribute values do not match.", expectedValue, actualValue);
      }

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testGetMetadataCurrent() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      String name = "testGetMetadataCurrent-" + getUniqueSuffix();

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      HashMap<String,String> attributes = getDefaultAttributes();
      options.setAttributes(attributes);

      String description = "test current description";
      options.setDescription(description);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion(name, material2);

      KeyProvider.Metadata actualMetadata = tkp.getMetadata(name);

      Assert.assertEquals("Ciphers do not match.", conf.get(KeyProvider.DEFAULT_CIPHER_NAME), actualMetadata.getCipher());
      Assert.assertTrue("Bit lengths do not match.", Integer.parseInt(conf.get(KeyProvider.DEFAULT_BITLENGTH_NAME)) == actualMetadata.getBitLength());
      Assert.assertEquals("Descriptions do not match.", description, actualMetadata.getDescription());

      Map<String, String> actualAttributes = actualMetadata.getAttributes();
      Assert.assertEquals("Attribute map sizes do not match.", attributes.size(), actualAttributes.size());

      for (String key : attributes.keySet()) {
        String expectedValue = attributes.get(key);
        String actualValue= actualAttributes.get(key);
        Assert.assertEquals("Attribute values do not match.", expectedValue, actualValue);
      }

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testDeleteKey() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testDeleteKey-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion(name, material2);

      tkp.deleteKey(name);

      KeyProvider.KeyVersion keyVersion = tkp.getCurrentKey(name);
      Assert.assertNull("getCurrentKey returned non-null value for deleted key.", keyVersion);

      try {
        keyVersion = tkp.getKeyVersion(keyVersion0.getVersionName());
        Assert.assertNull("getKeyVersion returned non-null value for deleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      }
      try {
        keyVersion = tkp.getKeyVersion(keyVersion1.getVersionName());
        Assert.assertNull("getKeyVersion returned non-null value for deleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      } try {
        keyVersion = tkp.getKeyVersion(keyVersion2.getVersionName());
        Assert.assertNull("getKeyVersion returned non-null value for deleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      }

      byte[] material3 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      try {
        tkp.createKey(name, material3, options);
        Assert.fail("Creation of key with same name (" + name + ") as deleted key succeeded when it should have failed.");
      } catch (TrusteeKeyProvider.DuplicateKeyException e) {
        // expected;
      }

     KeyProvider.Metadata metadata = tkp.getMetadata(name);
     Assert.assertNull("getMetadata returned non-null value for deleted key.", metadata);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testUnDeleteKey() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testUnDeleteKey-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.rollNewVersion(name, material1);
      KeyProvider.KeyVersion keyVersion2 = tkp.rollNewVersion(name, material2);

      tkp.deleteKey(name);
      tkp.unDeleteKey(name);

      KeyProvider.KeyVersion keyVersion = tkp.getCurrentKey(name);
      Assert.assertNotNull("getCurrentKey returned null value for undeleted key.", keyVersion);

      try {
        keyVersion = tkp.getKeyVersion(keyVersion0.getVersionName());
        Assert.assertNotNull("getKeyVersion returned null value for undeleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      }
      try {
        keyVersion = tkp.getKeyVersion(keyVersion1.getVersionName());
        Assert.assertNotNull("getKeyVersion returned null value for undeleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      } try {
        keyVersion = tkp.getKeyVersion(keyVersion2.getVersionName());
        Assert.assertNotNull("getKeyVersion returned null value for undeleted key.", keyVersion);
      } catch (TrusteeKeyProvider.KeyNotFoundException e) {
        // expected;
      }

      byte[] material3 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      try {
        tkp.createKey(name, material3, options);
        Assert.fail("Creation of key with same name (" + name + ") as undeleted key failed when it should have succeeded.");
      } catch (TrusteeKeyProvider.DuplicateKeyException e) {
        // expected;
      }

      KeyProvider.Metadata metadata = tkp.getMetadata(name);
      Assert.assertNotNull("getMetadata returned null value for undeleted key.", metadata);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testUnDeleteKeyMultipleClients() {
    TrusteeKeyProvider tkp1 = registerNewProvider();
    TrusteeKeyProvider tkp2 = registerNewProvider();

    try {
      String name = "testUnDeleteKeyMultipleClients-" + getUniqueSuffix();
      byte[] material00 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material01 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material02 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      byte[] material03 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material04 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material05 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp1.createKey(name, material00, options);
      KeyProvider.KeyVersion keyVersion1 = tkp1.rollNewVersion(name, material01);
      KeyProvider.KeyVersion keyVersion2 = tkp1.rollNewVersion(name, material02);

      KeyProvider.KeyVersion keyVersion3 = tkp2.createKey(name, material03, options);
      KeyProvider.KeyVersion keyVersion4 = tkp2.rollNewVersion(name, material04);
      KeyProvider.KeyVersion keyVersion5 = tkp2.rollNewVersion(name, material05);

      tkp1.deleteKey(name);

      KeyProvider.KeyVersion keyVersionDeleted = tkp1.getCurrentKey(name);
      Assert.assertNull("getCurrentKey returned non-null value for deleted key.", keyVersionDeleted);

      KeyProvider.KeyVersion keyVersionNotDeleted = tkp2.getCurrentKey(name);
      Assert.assertNotNull("getCurrentKey returned null value for not deleted key.", keyVersionNotDeleted);

      tkp2.deleteKey(name);

      tkp1.unDeleteKey(name);

      KeyProvider.KeyVersion keyVersionNowDeleted = tkp2.getCurrentKey(name);
      Assert.assertNull("getCurrentKey returned non-null value for deleted key.", keyVersionNowDeleted);

      KeyProvider.KeyVersion keyVersionUnDeleted = tkp1.getCurrentKey(name);
      Assert.assertNotNull("getCurrentKey returned null value for undeleted key.", keyVersionUnDeleted);

      KeyProvider.Metadata metadata = tkp1.getMetadata(name);
      Assert.assertNotNull("getMetadata returned null value for undeleted key.", metadata);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testPurgeKeysWithSimilarNames() {
    TrusteeKeyProvider tkp = registerNewProvider();

    try {
      String name1 = "testPurgeKeyWithSimilarNames-" + getUniqueSuffix();
      String name2 = name1 + "1";
      String name3 = name1 + "#!#$%";

      byte[] material00 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material01 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material02 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name1, material00, options);
      KeyProvider.KeyVersion keyVersion1 = tkp.createKey(name2, material01, options);
      KeyProvider.KeyVersion keyVersion2 = tkp.createKey(name3, material02, options);

      tkp.purgeKey(name1);

      KeyProvider.KeyVersion keyVersionPurged = tkp.getCurrentKey(name1);
      KeyProvider.KeyVersion keyVersionNotPurged1 = tkp.getCurrentKey(name2);
      KeyProvider.KeyVersion keyVersionNotPurged2 = tkp.getCurrentKey(name3);

      Assert.assertNull("getCurrentKey returned non-null value for purged key.", keyVersionPurged);
      Assert.assertNotNull("getCurrentKey returned null value for key with similar name to purged key", keyVersionNotPurged1);
      Assert.assertNotNull("getCurrentKey returned null value for key with similar name to purged key", keyVersionNotPurged2);

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  private HashMap<String, String> getDefaultAttributes() {
    HashMap<String, String> attributes = new HashMap<String, String>();
    String groupName = "group_name-" + getUniqueSuffix();
    attributes.put("group_name", groupName);
    return attributes;
  }

  @Test
  public void testMigration() {

    String[] hadoopConfigurations =  { "kms-site-migrate.xml" };
    TrusteeKeyProvider tkp = registerNewProvider(hadoopConfigurations);

    try {

      java.nio.file.Path currentPath = java.nio.file.Paths.get("");
      String currentAbsolutePath = currentPath.toAbsolutePath().toString();
      //System.out.println(currentAbsolutePath);
      URI uri = new URI("jceks://file@/" + currentAbsolutePath + "/src/test/resources/com.cloudera.keytrustee.test/kms.keystore");
      //System.out.println(uri);

      Configuration configuration = new Configuration(false); // don't load any default configuration files
      for (String resource : hadoopConfigurations) {
        configuration.addResource(resource);
      }
      KeyProvider sourceProvider = KeyProviderFactory.get(uri, configuration);
      List<String> keys = sourceProvider.getKeys();

      /*
      for (String key_name : keys) {
        System.out.println(key_name);
      }
      */

      for (String keyName : keys) {
        KeyProvider.KeyVersion sourceKeyVersion = sourceProvider.getCurrentKey(keyName);
        KeyProvider.Metadata sourceMetadata = sourceProvider.getMetadata(keyName);
        byte[] sourceKeyMaterial = sourceKeyVersion.getMaterial();
        String sourceKeyVersionName = sourceKeyVersion.getVersionName();

        KeyProvider.KeyVersion targetKeyVersionBefore = tkp.getCurrentKey(keyName);
        Assert.assertNull("Target provider should not contain key named " + keyName + " before migration.", targetKeyVersionBefore);
      }

      tkp.migrateKeys(sourceProvider);

      for (String keyName : keys) {
        KeyProvider.KeyVersion sourceKeyVersion = sourceProvider.getCurrentKey(keyName);
        KeyProvider.Metadata sourceMetadata = sourceProvider.getMetadata(keyName);
        byte[] sourceKeyMaterial = sourceKeyVersion.getMaterial();
        String sourceKeyVersionName = sourceKeyVersion.getVersionName();

        KeyProvider.KeyVersion targetKeyVersionAfter = tkp.getCurrentKey(keyName);
        Assert.assertNotNull("Target provider should contain key named " + keyName + " after migration.", targetKeyVersionAfter);
        Assert.assertArrayEquals("Key material in source and target do not match.", sourceKeyMaterial, targetKeyVersionAfter.getMaterial());

        KeyProvider.Metadata targetMetadata = tkp.getMetadata(keyName);
        String metadataSuffix = " metadata in source and target does not match.";
        Assert.assertEquals("Algorithm" + metadataSuffix, sourceMetadata.getAlgorithm(), targetMetadata.getAlgorithm());
        Assert.assertEquals("Cipher" + metadataSuffix, sourceMetadata.getCipher(), targetMetadata.getCipher());
        Assert.assertEquals("Description" + metadataSuffix, sourceMetadata.getDescription(), targetMetadata.getDescription());
        Assert.assertEquals("Bit length" + metadataSuffix, sourceMetadata.getBitLength(), targetMetadata.getBitLength());
        //Assert.assertEquals("Created" + metadataSuffix, 0, sourceMetadata.getCreated().compareTo(targetMetadata.getCreated()));
        Assert.assertEquals("Version count" + metadataSuffix, sourceMetadata.getVersions(), targetMetadata.getVersions());

        Map<String, String> attributes = targetMetadata.getAttributes();
        String originalId = attributes.get("original_id");
        Assert.assertNotNull("original_id attribute is null on migrated key. Lookup using legacy client data will fail.", originalId);
        Assert.assertEquals("original_id attribute on migrated key does not match original key version name in source.", sourceKeyVersionName, originalId);
      }

    } catch (IOException e) {
      throw new RuntimeException(e);
    } catch (URISyntaxException e) {
      throw new RuntimeException(e);
    }
  }

  @Test
  public void testConnectionReturnOnExceptionOrError() {
    TrusteeKeyProvider tkp = registerNewProvider();
    try {
      String name = "testConnectionPooling-" + getUniqueSuffix();
      byte[] material0 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material1 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);
      byte[] material2 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      Configuration conf = getDefaultConfiguration();
      KeyProvider.Options options = new KeyProvider.Options(conf);

      KeyProvider.KeyVersion keyVersion0 = tkp.createKey(name, material0, options);

      tkp.deleteKey(name);

      KeyProvider.KeyVersion keyVersion = tkp.getCurrentKey(name);
      Assert.assertNull("getCurrentKey returned non-null value for deleted key.", keyVersion);

      int initialActiveConnections = tkp.getActiveConnectionCount();

      for (int i=0; i < 30; i++) {
        try {
          keyVersion = tkp.getKeyVersion(keyVersion0.getVersionName());
          Assert.assertNull("getKeyVersion returned non-null value for deleted key.", keyVersion);
        } catch (TrusteeKeyProvider.KeyNotFoundException e) {
          Assert.assertEquals("Connection was not returned to pool on getKeyVersion failure iteration " + i + ".", initialActiveConnections, tkp.getActiveConnectionCount());
        }
      }

      byte[] material3 = getKeyMaterial(DEFAULT_ALGORITHM, DEFAULT_BITLENGTH);

      for (int i=0; i < 30; i++) {
        try {
          tkp.createKey(name, material3, options);
          Assert.fail("Creation of key with same name (" + name + ") as deleted key succeeded when it should have failed.");
        } catch (TrusteeKeyProvider.DuplicateKeyException e) {
          Assert.assertEquals("Connection was not returned to pool on createKey failure iteration " + i + ".", initialActiveConnections, tkp.getActiveConnectionCount());
        }
      }

    } catch (NoSuchAlgorithmException e) {
      throw new RuntimeException(e);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  /*
   * This test verifies that when SSL is activated in the Hadoop configuration (kts-site.xml) for an already
   * registered client, that the SSL connection is verified and the failure is reported immediately if the SSL
   * connection fails (because, for example, the client trust store does not contain the server certificate).
   */
  @Test
  public void testSSLUpdate(){

    try {

      java.net.URL configDirLongURL = getClass().getResource("/com.cloudera.keytrustee.test/keytrustee-cm-config-secure/");
      java.net.URL configDirShortURL = getClass().getResource("/keytrustee-cm-config-secure/");
      java.net.URL[] configDirCandidates = { configDirLongURL, configDirShortURL };
      String configPath = getValidPathString(configDirCandidates);

      if (null == configPath) {
        Assert.fail("Could not find keytrustee-cm-config-secure directory on classpath: " +
                System.getProperty("line.separator") +
                getClasspathString());
      }

      java.net.URL keyTrusteeConfLongURL = getClass().getResource("/com.cloudera.keytrustee.test/keytrustee-registered-insecure/keytrustee.conf");
      java.net.URL keyTrusteeConfShortURL = getClass().getResource("/keytrustee-registered-insecure/keytrustee.conf");
      java.net.URL[] keyTrusteeConfCandidates = { keyTrusteeConfLongURL, keyTrusteeConfShortURL };
      String confPath = getValidPathString(keyTrusteeConfCandidates);

      if (null == confPath) {
        Assert.fail("Could not find keytrustee-registered-insecure/keytrustee.conf file on classpath: " +
                System.getProperty("line.separator") +
                getClasspathString());
      }

      java.io.File tempFolder = folder.getRoot();
      java.io.File keyTrusteeConfDir = new java.io.File(confPath).getParentFile();
      org.apache.commons.io.FileUtils.copyDirectory(keyTrusteeConfDir, tempFolder, true);

      System.setProperty(TrusteeKeyProviderConfiguration.KTS_CONFIG_DIR, configPath);

      String[] hadoopConfigurations =  { "kts-site.xml" };

      // Next line should throw an exception since trust store does not contain server certificate when
      // SSL is activated.
      TrusteeKeyProvider tkp = registerOrLoadProvider(hadoopConfigurations, tempFolder.getAbsolutePath());
      Assert.fail("Expected javax.net.ssl.SSLHandshakeException did not occur.");
    }
    catch(Exception e)
    {
      Throwable cause = e.getCause();
      Assert.assertEquals(cause.getClass(), javax.net.ssl.SSLHandshakeException.class);
    } finally {
      System.clearProperty(TrusteeKeyProviderConfiguration.KTS_CONFIG_DIR);
    }
  }

  private String getValidPathString(java.net.URL[] urls) {
    String path = null;
    boolean fileExists = false;
    for (java.net.URL url : urls) {
      String testPath = url.getPath();
      java.io.File file = new java.io.File(testPath);
      if (file.exists()) {
        fileExists = true;
        path = testPath;
        break;
      }
    }
    return path;
  }

  private String getClasspathString() {
    //Get the System Classloader
    ClassLoader sysClassLoader = ClassLoader.getSystemClassLoader();

    //Get the URLs
    java.net.URL[] urls = ((java.net.URLClassLoader) sysClassLoader).getURLs();
    String classpath = "";

    for(int i=0; i< urls.length; i++)
    {
      classpath += ((String)urls[i].getFile() + System.getProperty("line.separator"));
    }
    return classpath;
  }
}
